import warnings
import keras 
import numpy as np


@keras.saving.register_keras_serializable(package='molcraft')
class GaussianNegativeLogLikelihood(keras.losses.Loss):

    def __init__(
        self, 
        events: int = 1, 
        name="gaussian_nll", 
        **kwargs
    ):
        super().__init__(name=name, **kwargs)
        self.events = events
    
    def call(self, y_true, y_pred):
        mean = y_pred[..., :self.events]
        scale = y_pred[..., self.events:]
        variance = keras.ops.square(scale)
        expected_rank = len(keras.ops.shape(mean))
        current_rank = len(keras.ops.shape(y_true))
        for _ in range(expected_rank - current_rank):
            y_true = keras.ops.expand_dims(y_true, axis=-1)
        return keras.ops.mean(
            0.5 * keras.ops.log(2.0 * np.pi * variance) + 
            0.5 * keras.ops.square(y_true - mean) / variance 
        )

    def get_config(self):
        config = super().get_config()
        config['events'] = self.events 
        return config 
    

GaussianNLL = GaussianNegativeLogLikelihood


import warnings
import keras 
import tensorflow as tf
import functools
import inspect
from keras.src.models import functional

from molcraft import tensors
from molcraft import ops 


@keras.saving.register_keras_serializable(package='molcraft')
class GraphLayer(keras.layers.Layer):
    """Base graph layer.

    Subclasses must implement a forward pass via **propagate(graph)**. 
    
    Subclasses may create dense layers and weights in **build(graph_spec)**.

    Note: `GraphLayer` currently only supports `GraphTensor` input.

    The list of arguments below are only relevant if the derived layer 
    invokes 'get_dense_kwargs`, `get_dense`  or `get_einsum_dense`. 

    Arguments:
        use_bias (bool):
            Whether bias should be used in dense layers. Default to `True`.
        kernel_initializer (keras.initializers.Initializer, str):
            Initializer for the kernel weight matrix of the dense layers.
            Default to `glorot_uniform`.
        bias_initializer (keras.initializers.Initializer, str):
            Initializer for the bias weight vector of the dense layers.
            Default to `zeros`.
        kernel_regularizer (keras.regularizers.Regularizer, None):
            Regularizer function applied to the kernel weight matrix.
            Default to `None`.
        bias_regularizer (keras.regularizers.Regularizer, None):
            Regularizer function applied to the bias weight vector.
            Default to `None`.
        activity_regularizer (keras.regularizers.Regularizer, None):
            Regularizer function applied to the output of the dense layers.
            Default to `None`.
        kernel_constraint (keras.constraints.Constraint, None):
            Constraint function applied to the kernel weight matrix.
            Default to `None`.
        bias_constraint (keras.constraints.Constraint, None):
            Constraint function applied to the bias weight vector.
            Default to `None`.
    """

    def __init__(
        self,
        use_bias: bool = True,
        kernel_initializer: keras.initializers.Initializer | str = "glorot_uniform",
        bias_initializer: keras.initializers.Initializer | str = "zeros",
        kernel_regularizer: keras.regularizers.Regularizer | None = None,
        bias_regularizer: keras.regularizers.Regularizer | None = None,
        activity_regularizer: keras.regularizers.Regularizer | None = None,
        kernel_constraint: keras.constraints.Constraint | None = None,
        bias_constraint: keras.constraints.Constraint | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self._use_bias = use_bias
        self._kernel_initializer = keras.initializers.get(kernel_initializer)
        self._bias_initializer = keras.initializers.get(bias_initializer)
        self._kernel_regularizer = keras.regularizers.get(kernel_regularizer)
        self._bias_regularizer = keras.regularizers.get(bias_regularizer)
        self._activity_regularizer = keras.regularizers.get(activity_regularizer)
        self._kernel_constraint = keras.constraints.get(kernel_constraint)
        self._bias_constraint = keras.constraints.get(bias_constraint)
        self._custom_build_config = {}
        self._propagate_kwargs = _propagate_kwargs(self.propagate)
        self.built = False

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        subclass_build = cls.build

        @functools.wraps(subclass_build)
        def build_wrapper(self: GraphLayer, spec: tensors.GraphTensor.Spec | None):
            GraphLayer.build(self, spec)
            subclass_build(self, spec)
            if not self.built and isinstance(self, keras.Model):
                symbolic_inputs = Input(spec)
                self.built = True
                self(symbolic_inputs)

        cls.build = build_wrapper

    def propagate(self, tensor: tensors.GraphTensor) -> tensors.GraphTensor:
        """Forward pass.

        Must be implemented by subclass.

        Arguments:
            tensor:
                A `GraphTensor` instance.
        """
        raise NotImplementedError(
            'The forward pass of the layer is not implemented. '
            'Please implement `propagate`.'
        )

    def build(self, tensor_spec: tensors.GraphTensor.Spec) -> None:
        """Builds the layer.

        May use built-in methods such as `get_weight`, `get_dense` and `get_einsum_dense`.

        Optionally implemented by subclass.

        Arguments:
            tensor_spec:
                A `GraphTensor.Spec` instance corresponding to the `GraphTensor` 
                passed to `propagate`.
        """ 
        if isinstance(tensor_spec, tensors.GraphTensor.Spec):
            self._custom_build_config['spec'] = _serialize_spec(tensor_spec)

    def call(
        self, 
        graph: dict[str, dict[str, tf.Tensor]],
        training: bool | None = None,
    ) -> dict[str, dict[str, tf.Tensor]]:
        graph_tensor = tensors.from_dict(graph)
        if not self._propagate_kwargs:
            outputs = self.propagate(graph_tensor)
        else:
            outputs = self.propagate(graph_tensor, training=training)
        if isinstance(outputs, tensors.GraphTensor):
            return tensors.to_dict(outputs)
        return outputs

    def __call__(
        self, 
        graph: dict[str, dict[str, tf.Tensor]] | tensors.GraphTensor, 
        **kwargs
    ) -> tf.Tensor | dict[str, dict[str, tf.Tensor]] | tensors.GraphTensor:
        if not self.built:
            spec = _spec_from_inputs(graph)
            self.build(spec)

        is_graph_tensor = isinstance(graph, tensors.GraphTensor)
        if is_graph_tensor:
            graph = tensors.to_dict(graph)
        else:
            graph = {field: dict(data) for (field, data) in graph.items()}

        if isinstance(self, functional.Functional):
            # As a functional model is strict for what input can be 
            # passed to it, we need to temporarily pop some of the 
            # input and add it back afterwards.
            symbolic_input = self._symbolic_input
            excluded = {}
            for outer_field in ['context', 'node', 'edge']:
                excluded[outer_field] = {}
                for inner_field in list(graph[outer_field]):
                    if inner_field not in symbolic_input[outer_field]:
                        excluded[outer_field][inner_field] = (
                            graph[outer_field].pop(inner_field)
                        )
            
            tf.nest.assert_same_structure(self.input, graph)

        outputs = super().__call__(graph, **kwargs)

        if not tensors.is_graph(outputs):
            return outputs
        
        graph = outputs
        if isinstance(self, functional.Functional):
            for outer_field in ['context', 'node', 'edge']:
                for inner_field in list(excluded[outer_field]):
                    graph[outer_field][inner_field] = (
                        excluded[outer_field].pop(inner_field)
                    )

        if is_graph_tensor:
            return tensors.from_dict(graph)

        return graph

    def get_build_config(self) -> dict:
        if self._custom_build_config:
            return self._custom_build_config
        return super().get_build_config()
    
    def build_from_config(self, config: dict) -> None:
        serialized_spec = config.get('spec')
        if serialized_spec is not None:
            spec = _deserialize_spec(serialized_spec)
            self.build(spec)
        else:
            super().build_from_config(config)

    def get_weight(
        self,
        shape: tf.TensorShape,
        **kwargs,
    ) -> tf.Variable:
        common_kwargs = self.get_dense_kwargs()
        weight_kwargs = {
            'initializer': common_kwargs['kernel_initializer'],
            'regularizer': common_kwargs['kernel_regularizer'],
            'constraint': common_kwargs['kernel_constraint']
        }
        weight_kwargs.update(kwargs)
        return self.add_weight(shape=shape, **weight_kwargs)
    
    def get_dense(
        self, 
        units: int, 
        **kwargs
    ) -> keras.layers.Dense:
        common_kwargs = self.get_dense_kwargs()
        common_kwargs.update(kwargs)
        return keras.layers.Dense(units, **common_kwargs)
    
    def get_einsum_dense(
        self, 
        equation: str, 
        output_shape: tf.TensorShape, 
        **kwargs
    ) -> keras.layers.EinsumDense:
        common_kwargs = self.get_dense_kwargs()
        common_kwargs.update(kwargs)
        use_bias = common_kwargs.pop('use_bias', False)
        if use_bias and not 'bias_axes' in common_kwargs:
            common_kwargs['bias_axes'] = equation.split('->')[-1][1:] or None
        return keras.layers.EinsumDense(equation, output_shape, **common_kwargs)
    
    def get_dense_kwargs(self) -> dict:
        common_kwargs = dict(
            use_bias=self._use_bias,
            kernel_regularizer=self._kernel_regularizer,
            bias_regularizer=self._bias_regularizer,
            activity_regularizer=self._activity_regularizer,
            kernel_constraint=self._kernel_constraint,
            bias_constraint=self._bias_constraint,
        )
        kernel_initializer = self._kernel_initializer.__class__.from_config(
            self._kernel_initializer.get_config()
        )
        bias_initializer = self._bias_initializer.__class__.from_config(
            self._bias_initializer.get_config()
        )
        common_kwargs["kernel_initializer"] = kernel_initializer
        common_kwargs["bias_initializer"] = bias_initializer
        return common_kwargs
    
    def get_config(self) -> dict:
        config = super().get_config()
        config.update({
            "use_bias": self._use_bias,
            "kernel_initializer": 
                keras.initializers.serialize(self._kernel_initializer),
            "bias_initializer": 
                keras.initializers.serialize(self._bias_initializer),
            "kernel_regularizer": 
                keras.regularizers.serialize(self._kernel_regularizer),
            "bias_regularizer": 
                keras.regularizers.serialize(self._bias_regularizer),
            "activity_regularizer":
                keras.regularizers.serialize(self._activity_regularizer),
            "kernel_constraint": 
                keras.constraints.serialize(self._kernel_constraint),
            "bias_constraint": 
                keras.constraints.serialize(self._bias_constraint),
        })
        return config

    @property
    def _symbolic_output(self) -> dict[dict[str, keras.KerasTensor]]:
        output = self.output
        if tensors.is_graph(output):
            # GraphModel
            return output
        symbolic_input = self._symbolic_input
        symbolic_output = self.compute_output_spec(symbolic_input)
        return tf.nest.pack_sequence_as(symbolic_output, output)

    @property
    def _symbolic_input(self) -> dict[dict[str, keras.KerasTensor]]:
        input = self.input
        if tensors.is_graph(input):
            # GraphModel or initial GraphLayer of GraphModel
            return input
        spec = _deserialize_spec(self.get_build_config()['spec'])
        spec_dict = {k: dict(v) for k, v in spec.__dict__.items()}
        return tf.nest.pack_sequence_as(spec_dict, input)

    @property
    def output_spec(self) -> tensors.GraphTensor.Spec | tf.TensorSpec:
        if not self.built:
            return None
        if isinstance(self, functional.Functional):
            output_spec = self.output
        else:
            serialized_spec = self.get_build_config()['spec']
            deserialized_spec = _deserialize_spec(serialized_spec)
            input_spec = Input(deserialized_spec)
            output_spec = self.compute_output_spec(input_spec)
        if not tensors.is_graph(output_spec):
            return tf.TensorSpec(output_spec.shape, output_spec.dtype)
        spec_dict = tf.nest.map_structure(
            lambda t: tf.TensorSpec(t.shape, t.dtype), output_spec
        )
        return tensors.GraphTensor.Spec(**spec_dict)


@keras.saving.register_keras_serializable(package='molcraft')
class GraphConv(GraphLayer):

    """Base graph neural network layer.

    This layer implements the three basic steps of a graph neural network layer, each of which 
    can (optionally) be overridden by the `GraphConv` subclass:

    1. **message(graph)**, which computes the *messages* to be passed to target nodes;
    2. **aggregate(graph)**, which *aggregates* messages to target nodes;
    3. **update(graph)**, which further *updates* (target) nodes.

    Note: for skip connection to work, the `GraphConv` subclass requires final node feature 
    output dimension to be equal to `units`. 

    Arguments:
        units (int):
            Dimensionality of the output space.
        activation (keras.layers.Activation, str, None):
            Activation function to be accessed via `self.activation`, and used for the 
            `message()` and `update()` methods, if not overriden. Default to `relu`.
        use_bias (bool):
            Whether bias should be used in the dense layers. Default to `True`.
        normalize (bool, str):
            Whether a normalization layer should be obtain by `get_norm()`. Default to `False`.
        skip_connect (bool):
            Whether node feature input should be added to the node feature output. Default to `True`.
        kernel_initializer (keras.initializers.Initializer, str):
            Initializer for the kernel weight matrix of the dense layers.
            Default to `glorot_uniform`.
        bias_initializer (keras.initializers.Initializer, str):
            Initializer for the bias weight vector of the dense layers.
            Default to `zeros`.
        kernel_regularizer (keras.regularizers.Regularizer, None):
            Regularizer function applied to the kernel weight matrix.
            Default to `None`.
        bias_regularizer (keras.regularizers.Regularizer, None):
            Regularizer function applied to the bias weight vector.
            Default to `None`.
        activity_regularizer (keras.regularizers.Regularizer, None):
            Regularizer function applied to the output of the dense layers.
            Default to `None`.
        kernel_constraint (keras.constraints.Constraint, None):
            Constraint function applied to the kernel weight matrix.
            Default to `None`.
        bias_constraint (keras.constraints.Constraint, None):
            Constraint function applied to the bias weight vector.
            Default to `None`.
    """
        
    def __init__(
        self, 
        units: int = None, 
        activation: str | keras.layers.Activation | None = 'relu',
        use_bias: bool = True,
        normalize: bool = False,
        skip_connect: bool = True,
        **kwargs
    ) -> None:
        super().__init__(use_bias=use_bias, **kwargs)
        self._units = units
        self._normalize = normalize
        self._skip_connect = skip_connect
        self._activation = keras.activations.get(activation)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        subclass_build = cls.build

        @functools.wraps(subclass_build)
        def build_wrapper(self, spec):
            GraphConv.build(self, spec)
            return subclass_build(self, spec)

        cls.build = build_wrapper

    @property 
    def units(self):
        return self._units 
    
    @property
    def activation(self):
        return self._activation
    
    def build(self, spec: tensors.GraphTensor.Spec) -> None:
        if not self.units:
            raise ValueError(
                f'`self.units` needs to be a positive integer. Found: {self.units}.'
            )
        node_feature_dim = spec.node['feature'].shape[-1]
        self._project_residual = (
            self._skip_connect and (node_feature_dim != self.units)
        )
        if self._project_residual:
            warnings.warn(
                'Found incompatible dim between input and output. Applying '
                'a projection layer to residual to match input and output dim.',
            )
            self._residual_dense = self.get_dense(
                self.units, name='residual_dense'
            )

        self.has_edge_feature = 'feature' in spec.edge
        self.has_node_coordinate = 'coordinate' in spec.node

        has_overridden_message = self.__class__.message != GraphConv.message 
        if not has_overridden_message:
            self._message_intermediate_dense = self.get_dense(self.units)
            self._message_norm = self.get_norm()
            self._message_intermediate_activation = self.activation
            self._message_final_dense = self.get_dense(self.units)

        has_overridden_aggregate = self.__class__.message != GraphConv.aggregate
        if not has_overridden_aggregate:
            pass

        has_overridden_update = self.__class__.update != GraphConv.update 
        if not has_overridden_update:
            self._update_intermediate_dense = self.get_dense(self.units)
            self._update_norm = self.get_norm()
            self._update_intermediate_activation = self.activation
            self._update_final_dense = self.get_dense(self.units)

    def propagate(self, tensor: tensors.GraphTensor) -> tensors.GraphTensor:
        """Forward pass.

        Invokes `message(graph)`, `aggregate(graph)` and `update(graph)` in sequence.

        Arguments:
            tensor:
                A `GraphTensor` instance.
        """
        if self._skip_connect:
            residual = tensor.node['feature']
            if self._project_residual:
                residual = self._residual_dense(residual)

        message = self.message(tensor)
        add_message = not isinstance(message, tensors.GraphTensor)
        if add_message:
            message = tensor.update({'edge': {'message': message}})
        elif not 'message' in message.edge:
            raise ValueError('Could not find `message` in `edge` output.')
        
        aggregate = self.aggregate(message)
        add_aggregate = not isinstance(aggregate, tensors.GraphTensor)
        if add_aggregate:
            aggregate = tensor.update({'node': {'aggregate': aggregate}})
        elif not 'aggregate' in aggregate.node:
            raise ValueError('Could not find `aggregate` in `node` output.')
            
        update = self.update(aggregate)
        if not isinstance(update, tensors.GraphTensor):
            update = tensor.update({'node': {'feature': update}})
        elif not 'feature' in update.node:
            raise ValueError('Could not find `feature` in `node` output.')
        
        if update.node['feature'].shape[-1] != self.units:
            raise ValueError('Updated node `feature` is not equal to `self.units`.')

        if add_message and add_aggregate:
            update = update.update({'node': {'aggregate': None}, 'edge': {'message': None}})
        elif add_message:
            update = update.update({'edge': {'message': None}})
        elif add_aggregate:
            update = update.update({'node': {'aggregate': None}})

        if not self._skip_connect:
            return update
        
        feature = update.node['feature']

        if self._skip_connect:
            feature += residual 

        return update.update({'node': {'feature': feature}})

    def message(self, tensor: tensors.GraphTensor) -> tensors.GraphTensor:
        """Compute messages.

        This method may be overridden by subclass.

        Arguments:
            tensor:
                The inputted `GraphTensor` instance.
        """
        message = keras.ops.concatenate(
            [
                tensor.gather('feature', 'source'),
                tensor.gather('feature', 'target'),
            ],
            axis=-1
        )
        if self.has_edge_feature:
            message = keras.ops.concatenate(
                [
                    message, 
                    tensor.edge['feature']
                ], 
                axis=-1
            )
        if self.has_node_coordinate:
            euclidean_distance = ops.euclidean_distance(
                tensor.gather('coordinate', 'target'),
                tensor.gather('coordinate', 'source'),
                axis=-1,
                keepdims=True
            )
            message = keras.ops.concatenate(
                [
                    message,
                    euclidean_distance
                ],
                axis=-1
            )
        message = self._message_intermediate_dense(message)
        message = self._message_norm(message)
        message = self._message_intermediate_activation(message)
        message = self._message_final_dense(message)
        return tensor.update({'edge': {'message': message}})

    def aggregate(self, tensor: tensors.GraphTensor) -> tensors.GraphTensor:
        """Aggregates messages.

        This method may be overridden by subclass.

        Arguments:
            tensor:
                A `GraphTensor` instance containing a message.
        """
        previous = tensor.node['feature']
        aggregate = tensor.aggregate('message', mode='mean')
        aggregate = keras.ops.concatenate([aggregate, previous], axis=-1)
        return tensor.update(
            {
                'node': {
                    'aggregate': aggregate, 
                },
                'edge': {
                    'message': None,
                }
            }
        )

    def update(self, tensor: tensors.GraphTensor) -> tensors.GraphTensor:
        """Updates nodes. 

        This method may be overridden by subclass.

        Arguments:
            tensor:
                A `GraphTensor` instance containing aggregated messages 
                (updated node features).
        """
        aggregate = tensor.node['aggregate']
        node_feature = self._update_intermediate_dense(aggregate)
        node_feature = self._update_norm(node_feature)
        node_feature = self._update_intermediate_activation(node_feature)
        node_feature = self._update_final_dense(node_feature)
        return tensor.update(
            {
                'node': {
                    'feature': node_feature,
                    'aggregate': None,
                },
            }
        )

    def get_norm(self, **kwargs):
        if not self._normalize:
            return keras.layers.Identity()
        elif str(self._normalize).lower().startswith('layer'):
            return keras.layers.LayerNormalization(**kwargs)
        else:
            return keras.layers.BatchNormalization(**kwargs)
        
    def get_config(self) -> dict:
        config = super().get_config()
        config.update({
            'units': self.units,
            'activation': keras.activations.serialize(self._activation),
            'normalize': self._normalize,
            'skip_connect': self._skip_connect,
        })
        return config
    

@keras.saving.register_keras_serializable(package='molcraft')
class GIConv(GraphConv):

    """Graph isomorphism network layer.

    >>> graph = molcraft.tensors.GraphTensor(
    ...     context={
    ...         'size': [2]
    ...     },
    ...     node={
    ...         'feature': [[1.], [2.]]
    ...     },
    ...     edge={
    ...         'source': [0, 1],
    ...         'target': [1, 0],
    ...     }
    ... )
    >>> conv = molcraft.layers.GIConv(units=4)
    >>> conv(graph)
        GraphTensor(
            context={
                'size': <tf.Tensor: shape=[1], dtype=int32>
            },
            node={
                'feature': <tf.Tensor: shape=[2, 4], dtype=float32>
            },
            edge={
                'source': <tf.Tensor: shape=[2], dtype=int32>,
                'target': <tf.Tensor: shape=[2], dtype=int32>
            }
        )
    """

    def __init__(
        self,
        units: int,
        activation: keras.layers.Activation | str | None = 'relu',
        use_bias: bool = True,
        normalize: bool = False,
        skip_connect: bool = True,
        update_edge_feature: bool = True,
        **kwargs,
    ):
        super().__init__(
            units=units, 
            activation=activation,
            use_bias=use_bias, 
            normalize=normalize, 
            skip_connect=skip_connect,
            **kwargs
        )
        self._update_edge_feature = update_edge_feature

    def build(self, spec: tensors.GraphTensor.Spec) -> None:
        node_feature_dim = spec.node['feature'].shape[-1]

        self.epsilon = self.add_weight(
            name='epsilon', 
            shape=(), 
            initializer='zeros',
            trainable=True,
        )

        if self.has_edge_feature:
            edge_feature_dim = spec.edge['feature'].shape[-1]

            if not self._update_edge_feature:
                if (edge_feature_dim != node_feature_dim):
                    warnings.warn(
                        'Found edge and node feature dim to be incompatible. Applying a '
                        'projection layer to edge features to match the dim of the node features.',
                    )
                    self._update_edge_feature = True 

            if self._update_edge_feature:
                self._edge_dense = self.get_dense(node_feature_dim)
        else:
            self._update_edge_feature = False

    def message(self, tensor: tensors.GraphTensor) -> tensors.GraphTensor:
        message = tensor.gather('feature', 'source')
        edge_feature = tensor.edge.get('feature')
        if self._update_edge_feature:
            edge_feature = self._edge_dense(edge_feature)
        if self.has_edge_feature:
            message += edge_feature
            message = keras.ops.relu(message)
        return tensor.update(
            {
                'edge': {
                    'message': message,
                    'feature': edge_feature
                }
            }
        )

    def aggregate(self, tensor: tensors.GraphTensor) -> tensors.GraphTensor:
        node_feature = tensor.aggregate('message', mode='mean')
        node_feature += (1 + self.epsilon) * tensor.node['feature']
        return tensor.update(
            {
                'node': {
                    'aggregate': node_feature,
                },
                'edge': {
                    'message': None,
                }
            }
        )
    
    def get_config(self) -> dict:
        config = super().get_config()
        config.update({
            'update_edge_feature': self._update_edge_feature
        })
        return config


@keras.saving.register_keras_serializable(package='molcraft')
class GAConv(GraphConv):

    """Graph attention network layer.

    >>> graph = molcraft.tensors.GraphTensor(
    ...     context={
    ...         'size': [2]
    ...     },
    ...     node={
    ...         'feature': [[1.], [2.]]
    ...     },
    ...     edge={
    ...         'source': [0, 1],
    ...         'target': [1, 0],
    ...     }
    ... )
    >>> conv = molcraft.layers.GAConv(units=4, heads=2)
    >>> conv(graph)
        GraphTensor(
            context={
                'size': <tf.Tensor: shape=[1], dtype=int32>
            },
            node={
                'feature': <tf.Tensor: shape=[2, 4], dtype=float32>
            },
            edge={
                'source': <tf.Tensor: shape=[2], dtype=int32>,
                'target': <tf.Tensor: shape=[2], dtype=int32>
            }
        )
    """

    def __init__(
        self,
        units: int,
        heads: int = 8,
        activation: keras.layers.Activation | str | None = "relu",
        use_bias: bool = True,
        normalize: bool = False,
        skip_connect: bool = True,
        update_edge_feature: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(
            units=units, 
            activation=activation,
            normalize=normalize, 
            use_bias=use_bias,
            skip_connect=skip_connect,
            **kwargs
        )
        self._heads = heads
        if self.units % self.heads != 0:
            raise ValueError(f"units need to be divisible by heads.")
        self._head_units = self.units // self.heads 
        self._update_edge_feature = update_edge_feature

    @property 
    def heads(self):
        return self._heads 
    
    @property 
    def head_units(self):
        return self._head_units 

    def build(self, spec: tensors.GraphTensor.Spec) -> None:
        self._update_edge_feature = self.has_edge_feature and self._update_edge_feature
        if self._update_edge_feature:
            self._edge_dense = self.get_einsum_dense(
                'ijh,jkh->ikh', (self.head_units, self.heads)
            )
        self._node_dense = self.get_einsum_dense(
            'ij,jkh->ikh', (self.head_units, self.heads)
        )
        self._feature_dense = self.get_einsum_dense(
            'ij,jkh->ikh', (self.head_units, self.heads)
        )
        self._attention_dense = self.get_einsum_dense(
            'ijh,jkh->ikh', (1, self.heads)
        )

    def message(self, tensor: tensors.GraphTensor) -> tensors.GraphTensor:
        attention_feature = keras.ops.concatenate(
            [
                tensor.gather('feature', 'source'),
                tensor.gather('feature', 'target')
            ], 
            axis=-1
        )
        if self.has_edge_feature:
            attention_feature = keras.ops.concatenate(
                [
                    attention_feature, 
                    tensor.edge['feature']
                ], 
                axis=-1
            )

        attention_feature = self._feature_dense(attention_feature)

        edge_feature = tensor.edge.get('feature')

        if self._update_edge_feature:
            edge_feature = self._edge_dense(attention_feature)
            edge_feature = keras.ops.reshape(edge_feature, (-1, self.units))
        
        attention_feature = keras.ops.leaky_relu(attention_feature)
        attention_score = self._attention_dense(attention_feature)
        attention_score = ops.edge_softmax(
            score=attention_score, edge_target=tensor.edge['target']
        )
        node_feature = self._node_dense(tensor.node['feature'])
        message = ops.gather(node_feature, tensor.edge['source'])
        message = ops.edge_weight(message, attention_score)
        return tensor.update(
            {
                'edge': {
                    'message': message,
                    'feature': edge_feature,
                }
            }
        )
    
    def aggregate(self, tensor: tensors.GraphTensor) -> tensors.GraphTensor:
        node_feature = tensor.aggregate('message', mode='sum')
        node_feature = keras.ops.reshape(node_feature, (-1, self.units))
        return tensor.update(
            {
                'node': {
                    'aggregate': node_feature
                },
                'edge': {
                    'message': None,
                }
            }
        )
    
    def get_config(self) -> dict:
        config = super().get_config()
        config.update({
            "heads": self._heads,
            'update_edge_feature': self._update_edge_feature,
        })
        return config
    

@keras.saving.register_keras_serializable(package='molcraft')
class MPConv(GraphConv):

    """Message passing neural network layer.

    Also supports 3D molecular graphs.

    >>> graph = molcraft.tensors.GraphTensor(
    ...     context={
    ...         'size': [2]
    ...     },
    ...     node={
    ...         'feature': [[1.], [2.]]
    ...     },
    ...     edge={
    ...         'source': [0, 1],
    ...         'target': [1, 0],
    ...     }
    ... )
    >>> conv = molcraft.layers.MPConv(units=4)
    >>> conv(graph)
        GraphTensor(
            context={
                'size': <tf.Tensor: shape=[1], dtype=int32>
            },
            node={
                'feature': <tf.Tensor: shape=[2, 4], dtype=float32>
            },
            edge={
                'source': <tf.Tensor: shape=[2], dtype=int32>,
                'target': <tf.Tensor: shape=[2], dtype=int32>
            }
        )
    """

    def __init__(
        self, 
        units: int = 128, 
        activation: keras.layers.Activation | str | None = 'relu', 
        use_bias: bool = True,
        normalize: bool = False,
        skip_connect: bool = True,
        **kwargs
    ) -> None:
        super().__init__(
            units=units, 
            activation=activation,
            use_bias=use_bias,
            normalize=normalize, 
            skip_connect=skip_connect,
            **kwargs
        )

    def build(self, spec: tensors.GraphTensor.Spec) -> None:
        node_feature_dim = spec.node['feature'].shape[-1]
        self.update_fn = keras.layers.GRUCell(self.units)
        self._project_previous_node_feature = node_feature_dim != self.units
        if self._project_previous_node_feature:
            warnings.warn(
                'Inputted node feature dim does not match updated node feature dim, '
                'which is required for the GRU update. Applying a projection layer to '
                'the inputted node features prior to the GRU update, to match dim '
                'of the updated node feature dim.'
            )
            self._previous_node_dense = self.get_dense(self.units)

    def aggregate(self, tensor: tensors.GraphTensor) -> tensors.GraphTensor:
        """Aggregates messages.

        This method may be overridden by subclass.

        Arguments:
            tensor:
                A `GraphTensor` instance containing a message.
        """
        aggregate = tensor.aggregate('message', mode='mean')
        return tensor.update(
            {
                'node': {
                    'aggregate': aggregate, 
                },
                'edge': {
                    'message': None,
                }
            }
        )
    
    def update(self, tensor: tensors.GraphTensor) -> tensors.GraphTensor:
        previous = tensor.node['feature']
        aggregate = tensor.node['aggregate']
        if self._project_previous_node_feature:
            previous = self._previous_node_dense(previous)
        updated_node_feature, _ = self.update_fn(
            inputs=aggregate, states=previous
        )
        return tensor.update(
            {
                'node': {
                    'feature': updated_node_feature,
                    'aggregate': None,
                }
            }
        )

    def get_config(self) -> dict:
        config = super().get_config()
        config.update({})
        return config 
    

@keras.saving.register_keras_serializable(package='molcraft')
class GTConv(GraphConv):

    """Graph transformer layer.

    Also supports 3D molecular graphs.

    >>> graph = molcraft.tensors.GraphTensor(
    ...     context={
    ...         'size': [2]
    ...     },
    ...     node={
    ...         'feature': [[1.], [2.]]
    ...     },
    ...     edge={
    ...         'source': [0, 1],
    ...         'target': [1, 0],
    ...     }
    ... )
    >>> conv = molcraft.layers.GTConv(units=4, heads=2)
    >>> conv(graph)
        GraphTensor(
            context={
                'size': <tf.Tensor: shape=[1], dtype=int32>
            },
            node={
                'feature': <tf.Tensor: shape=[2, 4], dtype=float32>
            },
            edge={
                'source': <tf.Tensor: shape=[2], dtype=int32>,
                'target': <tf.Tensor: shape=[2], dtype=int32>,
            }
        )
    """

    def __init__(
        self,
        units: int,
        heads: int = 8,
        activation: keras.layers.Activation | str | None = "relu",
        use_bias: bool = True,
        normalize: bool = False,
        skip_connect: bool = True,
        attention_dropout: float = 0.0,
        **kwargs,
    ) -> None:
        super().__init__(
            units=units, 
            activation=activation,
            normalize=normalize, 
            use_bias=use_bias,
            skip_connect=skip_connect,
            **kwargs
        )
        self._heads = heads
        if self.units % self.heads != 0:
            raise ValueError(f"units need to be divisible by heads.")
        self._head_units = self.units // self.heads 
        self._attention_dropout = attention_dropout

    @property 
    def heads(self):
        return self._heads 
    
    @property 
    def head_units(self):
        return self._head_units 
    
    def build(self, spec: tensors.GraphTensor.Spec) -> None:
        """Builds the layer.
        """
        self._query_dense = self.get_einsum_dense(
            'ij,jkh->ikh', (self.head_units, self.heads)
        )
        self._key_dense = self.get_einsum_dense(
            'ij,jkh->ikh', (self.head_units, self.heads)
        )
        self._value_dense = self.get_einsum_dense(
            'ij,jkh->ikh', (self.head_units, self.heads)
        )
        self._output_dense = self.get_dense(self.units)
        self._softmax_dropout = keras.layers.Dropout(self._attention_dropout) 

        if self.has_edge_feature:
            self._attention_bias_dense_1 = self.get_einsum_dense('ij,jkh->ikh', (1, self.heads))

        if self.has_node_coordinate:
            node_feature_dim = spec.node['feature'].shape[-1]
            num_kernels = self.units
            self._gaussian_loc = self.add_weight(
                shape=[num_kernels], initializer='zeros', dtype='float32', trainable=True
            ) 
            self._gaussian_scale = self.add_weight(
                shape=[num_kernels], initializer='ones', dtype='float32', trainable=True
            )
            self._centrality_dense = self.get_dense(units=node_feature_dim)
            self._attention_bias_dense_2 = self.get_einsum_dense('ij,jkh->ikh', (1, self.heads))

    def message(self, tensor: tensors.GraphTensor) -> tensors.GraphTensor:
        node_feature = tensor.node['feature']

        if self.has_node_coordinate:
            euclidean_distance = ops.euclidean_distance(
                tensor.gather('coordinate', 'target'),
                tensor.gather('coordinate', 'source'),
                axis=-1,
                keepdims=True
            )
            gaussian = ops.gaussian(
                euclidean_distance, self._gaussian_loc, self._gaussian_scale
            )
            centrality = keras.ops.segment_sum(gaussian, tensor.edge['target'], tensor.num_nodes)
            node_feature += self._centrality_dense(centrality)

        query = self._query_dense(node_feature)
        key = self._key_dense(node_feature)
        value = self._value_dense(node_feature)

        query = ops.gather(query, tensor.edge['source'])
        key = ops.gather(key, tensor.edge['target'])
        value = ops.gather(value, tensor.edge['source'])

        attention_score = keras.ops.sum(query * key, axis=1, keepdims=True)
        attention_score /= keras.ops.sqrt(float(self.head_units))
            
        if self.has_edge_feature:
            attention_score += self._attention_bias_dense_1(tensor.edge['feature'])

        if self.has_node_coordinate:
            attention_score += self._attention_bias_dense_2(gaussian)

        attention = ops.edge_softmax(attention_score, tensor.edge['target'])
        attention = self._softmax_dropout(attention)

        if self.has_node_coordinate:
            displacement = ops.displacement(
                tensor.gather('coordinate', 'target'),
                tensor.gather('coordinate', 'source'),
                normalize=True,
                axis=-1,
                keepdims=True,
            )
            attention *= keras.ops.expand_dims(displacement, axis=-1)
            attention = keras.ops.expand_dims(attention, axis=2)
            value = keras.ops.expand_dims(value, axis=1)

        message = ops.edge_weight(value, attention)
        
        return tensor.update(
            {
                'edge': {
                    'message': message,
                },
            }
        )

    def aggregate(self, tensor: tensors.GraphTensor) -> tensors.GraphTensor:
        node_feature = tensor.aggregate('message', mode='sum')
        if self.has_node_coordinate:
            shape = (tensor.num_nodes, -1, self.units)
        else:
            shape = (tensor.num_nodes, self.units)
        node_feature = keras.ops.reshape(node_feature, shape)
        node_feature = self._output_dense(node_feature)
        if self.has_node_coordinate:
            node_feature = keras.ops.sum(node_feature, axis=1)
        return tensor.update(
            {
                'node': {
                    'aggregate': node_feature,
                },
                'edge': {
                    'message': None,
                }
            }
        )
    
    def get_config(self) -> dict:
        config = super().get_config()
        config.update({
            "heads": self._heads,
            'attention_dropout': self._attention_dropout,
        })
        return config
    

@keras.saving.register_keras_serializable(package='molcraft')
class EGConv(GraphConv):

    """Equivariant graph neural network layer 3D.

    Only supports 3D molecular graphs.

    >>> graph = molcraft.tensors.GraphTensor(
    ...     context={
    ...         'size': [2]
    ...     },
    ...     node={
    ...         'feature': [[1.], [2.]],
    ...         'coordinate': [[0.1, -0.1, 0.5], [1.2, -0.5, 2.1]],
    ...     },
    ...     edge={
    ...         'source': [0, 1],
    ...         'target': [1, 0],
    ...     }
    ... )
    >>> conv = molcraft.layers.EGConv(units=4)
    >>> conv(graph)
        GraphTensor(
            context={
                'size': <tf.Tensor: shape=[1], dtype=int32>
            },
            node={
                'feature': <tf.Tensor: shape=[2, 4], dtype=float32>,
                'coordinate': <tf.Tensor: shape=[2, 3], dtype=float32>
            },
            edge={
                'source': <tf.Tensor: shape=[2], dtype=int32>,
                'target': <tf.Tensor: shape=[2], dtype=int32>
            }
        )
    """

    def __init__(
        self, 
        units: int = 128, 
        activation: keras.layers.Activation | str | None = 'silu', 
        use_bias: bool = True,
        normalize: bool = False,
        skip_connect: bool = True,
        **kwargs
    ) -> None:
        super().__init__(
            units=units, 
            activation=activation,
            use_bias=use_bias,
            normalize=normalize, 
            skip_connect=skip_connect,
            **kwargs
        )

    def build(self, spec: tensors.GraphTensor.Spec) -> None:
        if not self.has_node_coordinate:
            raise ValueError(
                'Could not find `coordinate`s in node, '
                'which is required for Conv3D layers.'
            )
        self._message_feedforward_intermediate = self.get_dense(
            self.units, activation=self.activation
        )
        self._message_feedforward_final = self.get_dense(
            self.units, activation=self.activation
        )

        self._coord_feedforward_intermediate = self.get_dense(
            self.units, activation=self.activation
        ) 
        self._coord_feedforward_final = self.get_dense(
            1, use_bias=False, activation='tanh'
        )

    def message(self, tensor: tensors.GraphTensor) -> tensors.GraphTensor:
        relative_node_coordinate = keras.ops.subtract(
            tensor.gather('coordinate', 'target'), 
            tensor.gather('coordinate', 'source')
        ) 
        squared_distance = keras.ops.sum(
            keras.ops.square(relative_node_coordinate), 
            axis=-1, 
            keepdims=True
        )
    
        # For numerical stability (i.e., to prevent NaN losses), this implementation of `EGConv3D` 
        # either needs to apply a `tanh` activation to the output of `self._coord_feedforward_final`, 
        # or normalize `relative_node_cordinate` as follows:
        #
        # norm = keras.ops.sqrt(squared_distance) + keras.backend.epsilon()
        # relative_node_coordinate /= norm
        #
        # For now, this implementation does the former.

        feature = keras.ops.concatenate(
            [
                tensor.gather('feature', 'target'), 
                tensor.gather('feature', 'source'), 
                squared_distance, 
            ], 
            axis=-1
        )
        if self.has_edge_feature:
            feature = keras.ops.concatenate(
                [
                    feature,
                    tensor.edge['feature']
                ], 
                axis=-1
            )
        message = self._message_feedforward_final(
            self._message_feedforward_intermediate(feature)
        )

        relative_node_coordinate = keras.ops.multiply(
            relative_node_coordinate, 
            self._coord_feedforward_final(
                self._coord_feedforward_intermediate(message)
            )
        )
        return tensor.update(
            {
                'edge': {
                    'message': message,
                    'relative_node_coordinate': relative_node_coordinate
                }
            }
        )

    def aggregate(self, tensor: tensors.GraphTensor) -> tensors.GraphTensor:
        coordinate = tensor.node['coordinate']
        coordinate += tensor.aggregate('relative_node_coordinate', mode='mean')

        # Original implementation seems to apply sum aggregation, which does not
        # seem work well for this implementation of `EGConv3D`, as it causes 
        # large output values and large initial losses. The magnitude of the 
        # aggregated values of a sum aggregation depends on the number of 
        # neighbors, which may be many and may differ from node to node (or 
        # graph to graph). Therefore, a mean mean aggregation is performed 
        # instead:
        aggregate = tensor.aggregate('message', mode='mean')
        aggregate = keras.ops.concatenate([aggregate, tensor.node['feature']], axis=-1)
        # Simply added to silence warning ('no gradients for variables ...')
        aggregate += (0.0 * keras.ops.sum(coordinate))

        return tensor.update(
            {
                'node': {
                    'aggregate': aggregate, 
                    'coordinate': coordinate,
                },
                'edge': {
                    'message': None,
                    'relative_node_coordinate': None
                }
            }
        ) 
    
    def get_config(self) -> dict:
        config = super().get_config()
        config.update({})
        return config 
    

@keras.saving.register_keras_serializable(package='molcraft')
class Readout(GraphLayer):

    """Readout layer.
    """

    def __init__(self, mode: str | None = None, **kwargs):
        kwargs['kernel_initializer'] = None 
        kwargs['bias_initializer'] = None
        super().__init__(**kwargs)
        self.mode = mode
        if str(self.mode).lower().startswith('sum'):
            self._reduce_fn = keras.ops.segment_sum
        elif str(self.mode).lower().startswith('max'):
            self._reduce_fn = keras.ops.segment_max 
        elif str(self.mode).lower().startswith('super'):
            self._reduce_fn = keras.ops.segment_sum
        else:
            self._reduce_fn = ops.segment_mean

    def propagate(self, tensor: tensors.GraphTensor) -> tf.Tensor:
        node_feature = tensor.node['feature']
        if str(self.mode).lower().startswith('super'):
            node_feature = keras.ops.where(
                tensor.node['super'][:, None], node_feature, 0.0
            )
        return self._reduce_fn(
            node_feature, tensor.graph_indicator, tensor.num_subgraphs
        )

    def get_config(self) -> dict:
        config = super().get_config()
        config['mode'] = self.mode 
        return config 
    

@keras.saving.register_keras_serializable(package='molcraft')
class NodeEmbedding(GraphLayer):

    """Node embedding layer.

    Embeds nodes based on its initial features.
    """

    def __init__(
        self, 
        dim: int | None = None, 
        intermediate_dim: int | None = None,
        intermediate_activation: str | keras.layers.Activation | None = 'relu',
        normalize: bool = False,
        embed_context: bool = False,
        num_wildcards: int | None = None,
        **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.dim = dim
        self._intermediate_dim = intermediate_dim
        self._intermediate_activation = keras.activations.get(
            intermediate_activation
        )
        self._normalize = normalize
        self._embed_context = embed_context
        self._num_wildcards = num_wildcards

    def build(self, spec: tensors.GraphTensor.Spec) -> None:
        feature_dim = spec.node['feature'].shape[-1]
        if not self.dim:
            self.dim = feature_dim
        if not self._intermediate_dim:
            self._intermediate_dim = self.dim * 2 
        self._node_dense = self.get_dense(
            self._intermediate_dim, activation=self._intermediate_activation
        )
        self._has_wildcard = 'wildcard' in spec.node
        self._has_super = 'super' in spec.node
        has_context_feature = 'feature' in spec.context
        if not has_context_feature:
            self._embed_context = False 
        if self._has_super and not self._embed_context:
            self._super_feature = self.get_weight(
                shape=[self._intermediate_dim], name='super_node_feature'
            )
        if self._embed_context:
            self._context_dense = self.get_dense(
                self._intermediate_dim, activation=self._intermediate_activation
            )
        if self._has_wildcard:
            if self._num_wildcards is None:
                warnings.warn(
                    "Found wildcards in input, but `num_wildcards` is set to `None`. "
                    "Automatically setting `num_wildcards` to 1. Please specify `num_wildcards>1` "
                    "if the layer should distinguish between different types of wildcards."
                )
                self._num_wildcards = 1
            self._wildcard_features = self.get_weight(
                shape=[self._num_wildcards, self._intermediate_dim], 
                name='wildcard_node_features'
            )
        if not self._normalize:
            self._norm = keras.layers.Identity()
        elif str(self._normalize).lower().startswith('layer'):
            self._norm = keras.layers.LayerNormalization()
        else:
            self._norm = keras.layers.BatchNormalization()
        self._dense = self.get_dense(self.dim)

    def propagate(self, tensor: tensors.GraphTensor) -> tensors.GraphTensor:
        feature = self._node_dense(tensor.node['feature'])

        if self._has_super and not self._embed_context:
            super_mask = keras.ops.expand_dims(tensor.node['super'], 1)
            super_feature = self._intermediate_activation(self._super_feature)
            feature = keras.ops.where(super_mask, super_feature, feature)

        if self._embed_context:
            context_feature = self._context_dense(tensor.context['feature'])
            feature = ops.scatter_update(feature, tensor.node['super'], context_feature)
            tensor = tensor.update({'context': {'feature': None}})

        if self._has_wildcard:
            wildcard = tensor.node['wildcard']
            wildcard_indices = keras.ops.where(wildcard > 0)[0]
            wildcard = ops.gather(wildcard, wildcard_indices)
            wildcard_feature = ops.gather(
                self._wildcard_features, keras.ops.mod(wildcard-1, self._num_wildcards)
            )
            wildcard_feature = self._intermediate_activation(wildcard_feature)
            feature = ops.scatter_update(feature, wildcard_indices, wildcard_feature)

        feature = self._norm(feature)
        feature = self._dense(feature)

        return tensor.update({'node': {'feature': feature}})

    def get_config(self) -> dict:
        config = super().get_config()
        config.update({
            'dim': self.dim,
            'intermediate_dim': self._intermediate_dim,
            'intermediate_activation': keras.activations.serialize(
                self._intermediate_activation
            ),
            'normalize': self._normalize,
            'embed_context': self._embed_context,
            'num_wildcards': self._num_wildcards,
        })
        return config
    

@keras.saving.register_keras_serializable(package='molcraft')
class EdgeEmbedding(GraphLayer):

    """Edge embedding layer.

    Embeds edges based on its initial features.
    """

    def __init__(
        self, 
        dim: int = None, 
        intermediate_dim: int | None = None,
        intermediate_activation: str | keras.layers.Activation | None = 'relu',
        normalize: bool = False,
        **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.dim = dim
        self._intermediate_dim = intermediate_dim
        self._intermediate_activation = keras.activations.get(
            intermediate_activation
        )
        self._normalize = normalize

    def build(self, spec: tensors.GraphTensor.Spec) -> None:
        feature_dim = spec.edge['feature'].shape[-1]
        if not self.dim:
            self.dim = feature_dim
        if not self._intermediate_dim:
            self._intermediate_dim = self.dim * 2 
        self._edge_dense = self.get_dense(
            self._intermediate_dim, activation=self._intermediate_activation
        ) 
        self._self_loop_feature = self.get_weight(
            shape=[self._intermediate_dim], name='self_loop_edge_feature'
        )
        self._has_super = 'super' in spec.edge
        if self._has_super:
            self._super_feature = self.get_weight(
                shape=[self._intermediate_dim], name='super_edge_feature'
            )
        if not self._normalize:
            self._norm = keras.layers.Identity()
        elif str(self._normalize).lower().startswith('layer'):
            self._norm = keras.layers.LayerNormalization()
        else:
            self._norm = keras.layers.BatchNormalization()
        self._dense = self.get_dense(self.dim)

    def propagate(self, tensor: tensors.GraphTensor) -> tensors.GraphTensor:
        feature = self._edge_dense(tensor.edge['feature'])

        if self._has_super:
            super_mask = keras.ops.expand_dims(tensor.edge['super'], 1)
            super_feature = self._intermediate_activation(self._super_feature)
            feature = keras.ops.where(super_mask, super_feature, feature)

        self_loop_mask = keras.ops.expand_dims(tensor.edge['source'] == tensor.edge['target'], 1)
        self_loop_feature = self._intermediate_activation(self._self_loop_feature)
        feature = keras.ops.where(self_loop_mask, self_loop_feature, feature)
        feature = self._norm(feature)
        feature = self._dense(feature)
        return tensor.update({'edge': {'feature': feature}})

    def get_config(self) -> dict:
        config = super().get_config()
        config.update({
            'dim': self.dim,
            'intermediate_dim': self._intermediate_dim,
            'intermediate_activation': keras.activations.serialize(
                self._intermediate_activation
            ),
            'normalize': self._normalize,
        })
        return config
    

@keras.saving.register_keras_serializable(package='molcraft')
class AddContext(GraphLayer):

    """Context adding layer.

    Adds context to super nodes or nodes, depending on whether super nodes exist.
    """

    def __init__(
        self, 
        field: str = 'feature',
        intermediate_dim: int | None = None,
        intermediate_activation: str | keras.layers.Activation | None = 'relu',
        drop: bool = False,
        normalize: bool = False,
        num_categories: int | None = None,
        **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self._field = field
        self._drop = drop
        self._intermediate_dim = intermediate_dim
        self._intermediate_activation = keras.activations.get(
            intermediate_activation
        )
        self._normalize = normalize
        self._num_categories = num_categories
        
    def build(self, spec: tensors.GraphTensor.Spec) -> None:
        is_categorical = spec.context[self._field].dtype.is_integer
        if is_categorical and not self._num_categories:
            raise ValueError(
                f'Found context ({self._field}) to be categorical (`int` dtype), but `num_categories`'
                'to be `None`. Please specify `num_categories` for categorical context.'
            )
        elif not is_categorical and self._num_categories:
            warnings.warn(
                f'`num_categories` is set to {self._num_categories}, but found context to be '
                'continuous (`float` dtype). Layer will cast context from `float` to `int` '
                'before one-hot encoding it.'
            )
        feature_dim = spec.node['feature'].shape[-1]
        self._has_super_node = 'super' in spec.node
        if self._intermediate_dim is None:
            self._intermediate_dim = feature_dim * 2
        self._intermediate_dense = self.get_dense(
            self._intermediate_dim, activation=self._intermediate_activation
        )
        self._final_dense = self.get_dense(feature_dim)
        if not self._normalize:
            self._intermediate_norm = keras.layers.Identity()
        elif str(self._normalize).lower().startswith('layer'):
            self._intermediate_norm = keras.layers.LayerNormalization()
        else:
            self._intermediate_norm = keras.layers.BatchNormalization()

    def propagate(self, tensor: tensors.GraphTensor) -> tensors.GraphTensor:
        context = tensor.context[self._field]
        if self._num_categories:
            if context.dtype.is_floating:
                context = keras.ops.cast(context, dtype=tensor.edge['source'].dtype)
            context = keras.utils.to_categorical(context, self._num_categories)
        elif len(keras.ops.shape(context)) == 1:
            context = keras.ops.expand_dims(context, axis=1)
        context = self._intermediate_dense(context)
        context = self._intermediate_norm(context)
        context = self._final_dense(context)
        if self._has_super_node:
            node_feature = ops.scatter_add(
                tensor.node['feature'], tensor.node['super'], context
            )
        else:
            node_feature = (
                tensor.node['feature'] + ops.gather(context, tensor.graph_indicator)
            )
        data = {'node': {'feature': node_feature}}
        if self._drop:
            data['context'] = {self._field: None}
        return tensor.update(data)

    def get_config(self) -> dict:
        config = super().get_config()
        config.update({
            'field': self._field,
            'intermediate_dim': self._intermediate_dim,
            'intermediate_activation': keras.activations.serialize(
                self._intermediate_activation
            ),
            'drop': self._drop,
            'normalize': self._normalize,
            'num_categories': self._num_categories,
        })
        return config


@keras.saving.register_keras_serializable(package='molcraft')
class GraphNetwork(GraphLayer):

    """Graph neural network.

    Sequentially calls graph layers (`GraphLayer`) and concatenates its output. 

    Arguments:
        layers (list):
            A list of graph layers.
    """

    def __init__(self, layers: list[GraphLayer], **kwargs) -> None:
        super().__init__(**kwargs)
        self.layers = layers
        self._update_edge_feature = False

    def build(self, spec: tensors.GraphTensor.Spec) -> None:
        units = self.layers[0].units 
        node_feature_dim = spec.node['feature'].shape[-1]
        self._update_node_feature = node_feature_dim != units 
        if self._update_node_feature:
            warnings.warn(
                'Node feature dim does not match `units` of the first layer. '
                'Applying a projection layer to node features to match `units`.',
            )
            self._node_dense = self.get_dense(units)
        self._has_edge_feature = 'feature' in spec.edge 
        if self._has_edge_feature:
            edge_feature_dim = spec.edge['feature'].shape[-1]
            self._update_edge_feature = edge_feature_dim != units
            if self._update_edge_feature:
                warnings.warn(
                    'Edge feature dim does not match `units` of the first layer. '
                    'Applying projection layer to edge features to match `units`.'
                )
                self._edge_dense = self.get_dense(units)

    def propagate(self, tensor: tensors.GraphTensor) -> tensors.GraphTensor:
        x = tensors.to_dict(tensor)
        if self._update_node_feature:
            x['node']['feature'] = self._node_dense(tensor.node['feature'])
        if self._has_edge_feature and self._update_edge_feature:
            x['edge']['feature'] = self._edge_dense(tensor.edge['feature'])
        outputs = [x['node']['feature']]
        for layer in self.layers:
            x = layer(x)
            outputs.append(x['node']['feature'])
        return tensor.update(
            {
                'node': {
                    'feature': keras.ops.concatenate(outputs, axis=-1)
                }    
            }
        )
    
    def tape_propagate(
        self,
        tensor: tensors.GraphTensor,
        tape: tf.GradientTape,
        training: bool | None = None,
    ) -> tuple[tensors.GraphTensor, list[tf.Tensor]]:
        """Performs the propagation with a `GradientTape`.

        Performs the same forward pass as `propagate` but with a `GradientTape`
        watching intermediate node features.

        Arguments:
            tensor (tensors.GraphTensor):
                The graph input.
        """
        if isinstance(tensor, tensors.GraphTensor):
            x = tensors.to_dict(tensor)
        else:
            x = tensor
        if self._update_node_feature:
            x['node']['feature'] = self._node_dense(tensor.node['feature'])
        if self._update_edge_feature:
            x['edge']['feature'] = self._edge_dense(tensor.edge['feature'])
        tape.watch(x['node']['feature'])
        outputs = [x['node']['feature']]
        for layer in self.layers:
            x = layer(x, training=training)
            tape.watch(x['node']['feature'])
            outputs.append(x['node']['feature'])

        tensor = tensor.update(
            {
                'node': {
                    'feature': keras.ops.concatenate(outputs, axis=-1)
                }
            }
        )
        return tensor, outputs
    
    def get_config(self) -> dict:
        config = super().get_config()
        config.update(
            {
                'layers': [
                    keras.layers.serialize(layer) for layer in self.layers
                ]
            }
        )
        return config

    @classmethod
    def from_config(cls, config: dict) -> 'GraphNetwork':
        config['layers'] = [
            keras.layers.deserialize(layer) for layer in config['layers']
        ]
        return super().from_config(config)
    

@keras.saving.register_keras_serializable(package='molcraft')
class GaussianParams(keras.layers.Dense):
    '''Gaussian parameters.

    Computes loc and scale via a dense layer. Should be implemented
    as the last layer in a model and paired with `losses.GaussianNLL`.

    The loc and scale parameters (resulting from this layer) are concatenated
    together along the last axis, resulting in a single output tensor. 

    Args:
        events (int):
            The number of events. If the model makes a single prediction per example,
            then the number of events should be 1. If the model makes multiple predictions 
            per example, then the number of events should be greater than 1. 
            Default to 1.
        kwargs:
            See `keras.layers.Dense` documentation. `activation` will be applied
            to `loc` only. `scale` is automatically softplus activated.
    '''
    def __init__(self, events: int = 1, **kwargs):
        units = kwargs.pop('units', None)
        activation = kwargs.pop('activation', None)
        if units:
            if units % 2 != 0:
                raise ValueError(
                    '`units` needs to be divisble by 2 as `units` = 2 x `events`.'
                )
        else:
            units = int(events * 2)
        super().__init__(units=units, **kwargs)
        self.events = events
        self.loc_activation = keras.activations.get(activation)

    def call(self, inputs, **kwargs):
        loc_and_scale = super().call(inputs, **kwargs)
        loc = loc_and_scale[..., :self.events]
        scale = loc_and_scale[..., self.events:]
        scale = keras.ops.softplus(scale) + keras.backend.epsilon()
        loc = self.loc_activation(loc)
        return keras.ops.concatenate([loc, scale], axis=-1)

    def get_config(self):
        config = super().get_config()
        config['events'] = self.events
        config['units'] = None
        config['activation'] = keras.activations.serialize(self.loc_activation)
        return config
    

def Input(spec: tensors.GraphTensor.Spec) -> dict:
    """Used to specify inputs to a functional model.

    Example:

    >>> import molcraft 
    >>> import keras
    >>> 
    >>> featurizer = molcraft.featurizers.MolGraphFeaturizer()
    >>> graph = featurizer([('N[C@@H](C)C(=O)O', 1.0), ('N[C@@H](CS)C(=O)O', 2.0)])
    >>> 
    >>> model = molcraft.models.GraphModel.from_layers(
    ...     molcraft.layers.Input(graph.spec),
    ...     molcraft.layers.NodeEmbedding(128),
    ...     molcraft.layers.EdgeEmbedding(128),
    ...     molcraft.layers.GraphTransformer(128),
    ...     molcraft.layers.GraphTransformer(128),
    ...     molcraft.layers.Readout('mean'),
    ...     molcraft.layers.Dense(1)
    ... ])
    """
    
    # Currently, Keras (3.8.0) does not support extension types.
    # So for now, this function will unpack the `GraphTensor.Spec` and 
    # return a dictionary of nested tensor specs. However, the corresponding 
    # nest of tensors will temporarily be converted to a `GraphTensor` by the 
    # `GraphLayer`, to levarage the utility of a `GraphTensor` object. 
    inputs = {}
    for outer_field, data in spec.__dict__.items():
        inputs[outer_field] = {}
        for inner_field, nested_spec in data.items():
            if inner_field in ['label', 'sample_weight']:
                # Remove label and sample_weight from the symbolic input as
                # a functional model is strict for what input can be passed.
                continue
            kwargs = {
                'shape': nested_spec.shape[1:],
                'dtype': nested_spec.dtype,
                'name': f'{outer_field}_{inner_field}'
            }
            if isinstance(nested_spec, tf.RaggedTensorSpec):
                # kwargs['ragged'] = True
                raise ValueError(
                    'Graph layers only supports graph input with nested `tf.Tensor` values.'
                )
            try:
                inputs[outer_field][inner_field] = keras.Input(**kwargs)
            except TypeError:
                raise ValueError(
                    "`keras.Input` does not currently support ragged tensors. For now, "
                    "pass the `Spec` of a 'flat' `GraphTensor` to `GNNInput`." 
                )
    return inputs


def _serialize_spec(spec: tensors.GraphTensor.Spec) -> dict:
    serialized_spec = {}
    for outer_field, data in spec.__dict__.items():
        serialized_spec[outer_field] = {}
        for inner_field, inner_spec in data.items():
            if inner_field in ['label', 'sample_weight']:
                continue
            serialized_spec[outer_field][inner_field] = {
                'shape': inner_spec.shape.as_list(), 
                'dtype': inner_spec.dtype.name, 
                'name': inner_spec.name,
            }
    return serialized_spec

def _deserialize_spec(serialized_spec: dict) -> tensors.GraphTensor.Spec:
    deserialized_spec = {}
    for outer_field, data in serialized_spec.items():
        deserialized_spec[outer_field] = {}
        for inner_field, inner_spec in data.items():
            deserialized_spec[outer_field][inner_field] = tf.TensorSpec(
                inner_spec['shape'], inner_spec['dtype'], inner_spec['name']
            )
    return tensors.GraphTensor.Spec(**deserialized_spec)

def _spec_from_inputs(inputs):
    symbolic_inputs = keras.backend.is_keras_tensor(
        tf.nest.flatten(inputs)[0]
    )
    if not symbolic_inputs:
        nested_specs = tf.nest.map_structure(
            tf.type_spec_from_value, inputs
        )
    else:
        nested_specs = tf.nest.map_structure(
            lambda t: tf.TensorSpec(t.shape, t.dtype), inputs
        )
    if isinstance(nested_specs, tensors.GraphTensor.Spec):
        spec = nested_specs
        return spec
    return tensors.GraphTensor.Spec(**nested_specs)

def _propagate_kwargs(func) -> bool:
    signature = inspect.signature(func)
    return any(
        (param.kind == inspect.Parameter.VAR_KEYWORD) or (param.name == 'training')
        for param in signature.parameters.values()
    )


import warnings
import numpy as np
import pandas as pd
import typing


def split(
    data: pd.DataFrame | np.ndarray,
    *,
    train_size: float | None = None,
    validation_size: float | None = None, 
    test_size: float | None = None,
    groups: str | np.ndarray = None,
    shuffle: bool = False, 
    random_seed: int | None = None,
) -> tuple[np.ndarray | pd.DataFrame, ...]:
    """Splits the dataset into subsets.

    Args:
        data: 
            A pd.DataFrame or np.ndarray object.
        train_size:
            The size of the train set.
        validation_size:
            The size of the validation set.
        test_size:
            The size of the test set.
        groups:
            The groups to perform the splitting on.
        shuffle:
            Whether the dataset should be shuffled prior to splitting.
        random_seed:
            The random state/seed. Only applicable if shuffling.
    """
    if not isinstance(data, (pd.DataFrame, np.ndarray)):
        raise ValueError(f'Unsupported `data` type ({type(data)}).')

    if isinstance(groups, str):
        groups = data[groups].values 
    elif groups is None:
        groups = np.arange(len(data))
    
    indices = np.unique(groups)
    size = len(indices)

    if not train_size and not test_size:
        raise ValueError(
            f'Found both `train_size` and `test_size` to be `None`, '
            f'specify at least one of them.'
        )
    if isinstance(test_size, float):
        test_size = int(size * test_size)
    if isinstance(train_size, float):
        train_size = int(size * train_size)
    if isinstance(validation_size, float):
        validation_size = int(size * validation_size)
    elif not validation_size:
        validation_size = 0

    if not train_size:
        train_size = (size - test_size - validation_size)
    if not test_size:
        test_size = (size - train_size - validation_size)
    
    remainder = size - (train_size + validation_size + test_size)
    if remainder < 0:
        raise ValueError(
            f'subset sizes added up to more than the data size.'
        )
    train_size += remainder

    if shuffle:
        np.random.seed(random_seed)
        np.random.shuffle(indices)

    train_mask = np.isin(groups, indices[:train_size])
    test_mask = np.isin(groups, indices[-test_size:])
    if not validation_size:
        return data[train_mask], data[test_mask]
    validation_mask = np.isin(groups, indices[train_size:-test_size])
    return data[train_mask], data[validation_mask], data[test_mask]
    
def cv_split(
    data: pd.DataFrame | np.ndarray,
    num_splits: int = 10,
    groups: str | np.ndarray = None,
    shuffle: bool = False, 
    random_seed: int | None = None,
) -> typing.Iterator[
        tuple[np.ndarray | pd.DataFrame, np.ndarray | pd.DataFrame]
    ]:
    """Splits the dataset into cross-validation folds.

    Args:
        data: 
            A pd.DataFrame or np.ndarray object.
        num_splits:
            The number of cross-validation folds.
        groups:
            The groups to perform the splitting on.
        shuffle:
            Whether the dataset should be shuffled prior to splitting.
        random_seed:
            The random state/seed. Only applicable if shuffling.
    """
    if not isinstance(data, (pd.DataFrame, np.ndarray)):
        raise ValueError(f'Unsupported `data` type ({type(data)}).')
    
    if isinstance(groups, str):
        groups = data[groups].values
    elif groups is None:
        groups = np.arange(len(data))

    indices = np.unique(groups)
    size = len(indices)
    
    if num_splits > size:
        raise ValueError(
            f'`num_splits` ({num_splits}) must not be greater than'
            f'the data size or the number of groups ({size}).'
        )
    if shuffle:
        np.random.seed(random_seed)
        np.random.shuffle(indices)

    indices_splits = np.array_split(indices, num_splits)

    for k in range(num_splits):
        test_indices = indices_splits[k]
        test_mask = np.isin(groups, test_indices)
        train_mask = ~test_mask
        yield data[train_mask], data[test_mask]



import warnings
import abc
import math
import keras
import numpy as np

from molcraft import chem


@keras.saving.register_keras_serializable(package='molcraft')
class Feature(abc.ABC):

    def __init__(
        self, 
        vocab: set[int | str] = None, 
        allow_oov: bool = True,
        encode_oov: bool = False, 
        dtype: str = 'float32'
    ) -> None:
        self.encode_oov = encode_oov
        self.allow_oov = allow_oov
        self.oov_token = '<oov>'
        self.dtype = dtype 
        if not vocab:
            vocab = default_vocabulary.get(self.name, None)
        if vocab:
            if isinstance(vocab, set):
                vocab: list = list(vocab)
                vocab.sort(key=lambda x: x if x is not None else "")
            elif not isinstance(vocab, list):
                vocab: list = list(vocab)
            if self.encode_oov and self.oov_token not in vocab:
                vocab.append(self.oov_token)
            onehot_encodings = np.eye(len(vocab), dtype=self.dtype)
            self.feature_to_onehot = dict(zip(vocab, onehot_encodings))
        self.vocab = vocab 

    @abc.abstractmethod
    def call(self, mol: chem.Mol) -> list[float | int | bool | str]: 
        pass
    
    def __call__(self, mol: chem.Mol) -> np.ndarray:
        if not isinstance(mol, chem.Mol):
            raise TypeError(f'Input to {self.name} must be a `chem.Mol` object.')
        features = self.call(mol)
        if len(features) != mol.num_atoms and len(features) != mol.num_bonds:
            raise ValueError(
                f'The number of features computed by {self.name} does not '
                'match the number of atoms or bonds of the `chem.Mol` object. '
                'Make sure to iterate over `atoms` or `bonds` of the `chem.Mol` '
                'object when computing features.'
            )
        if len(features) == 0:
            # Edge case: no atoms or bonds in the molecule.
            return np.zeros((0, self.output_dim), dtype=self.dtype)
        
        func = (
            self._featurize_categorical if self.vocab else 
            self._featurize_floating
        )
        return np.stack([func(x) for x in features])
        

    def get_config(self) -> dict:
        config = {
            'vocab': self.vocab,
            'allow_oov': self.allow_oov,
            'encode_oov': self.encode_oov,
            'dtype': self.dtype
        }
        return config
    
    @classmethod
    def from_config(cls, config: dict) -> 'Feature':
        return cls(**config)
    
    @property 
    def name(self) -> str:
        return self.__class__.__name__
    
    @property 
    def output_dim(self) -> int:
        return 1 if not self.vocab else len(self.vocab)
    
    def _featurize_categorical(self, feature: str | int) -> np.ndarray:
        encoding = self.feature_to_onehot.get(feature, None)
        if encoding is not None:
            return encoding
        if not self.allow_oov:
            raise ValueError(
                f'{feature} could not be encoded, as it was not found in `vocab`. '
                'To allow OOV features, set `allow_oov` or `encode_oov` to True.'
            )
        oov_encoding = self.feature_to_onehot.get(self.oov_token, None)
        if oov_encoding is None:
            oov_encoding = np.zeros([self.output_dim], dtype=self.dtype) 
        return oov_encoding
    
    def _featurize_floating(self, value: float | list[float]) -> np.ndarray:
        if not isinstance(value, (int, float, bool)):
            raise ValueError(
                f'{self.name} produced a value of type {type(value)}. ' 
                'If it represents a categorical feature, please provide a `vocab` '
                'to the constructor. If if represents a floating point feature, '
                'please make sure its `call` method returns a list of values of '
                'type `float`, `int`, `bool` or `None`.'
            )
        if not math.isfinite(value):
            warnings.warn(
                f'Found value of {self.name} to be non-finite. '
                f'Value received: {value}. Converting it to a value of 0.',
            )
            value = 0.0
        return np.asarray([value], dtype=self.dtype)
    

@keras.saving.register_keras_serializable(package='molcraft')
class AtomType(Feature):
    def call(self, mol: chem.Mol) -> list[int, float, str]: 
        return [atom.GetSymbol() for atom in mol.atoms]
    

@keras.saving.register_keras_serializable(package='molcraft')
class Degree(Feature):
    def call(self, mol: chem.Mol) -> list[int, float, str]: 
        return [atom.GetDegree() for atom in mol.atoms]


@keras.saving.register_keras_serializable(package='molcraft')
class NumHydrogens(Feature):
    def call(self, mol: chem.Mol) -> list[int, float, str]: 
        return [atom.GetTotalNumHs() for atom in mol.atoms]


@keras.saving.register_keras_serializable(package='molcraft')
class Valence(Feature):
    def call(self, mol: chem.Mol) -> list[int, float, str]: 
        return [atom.GetTotalValence() for atom in mol.atoms]
    

@keras.saving.register_keras_serializable(package='molcraft')
class AtomicWeight(Feature):
    def call(self, mol: chem.Mol) -> list[int, float, str]: 
        pt = chem.get_periodic_table()
        return [pt.GetAtomicWeight(atom.GetSymbol()) for atom in mol.atoms]
    

@keras.saving.register_keras_serializable(package='molcraft')
class Hybridization(Feature):
    def call(self, mol: chem.Mol) -> list[int, float, str]: 
        return [str(atom.GetHybridization()).lower() for atom in mol.atoms]


@keras.saving.register_keras_serializable(package='molcraft')
class CIPCode(Feature):
    def call(self, mol: chem.Mol) -> list[int, float, str]: 
        return [
            atom.GetProp("_CIPCode") if atom.HasProp("_CIPCode") else "None" 
            for atom in mol.atoms]
    

@keras.saving.register_keras_serializable(package='molcraft')
class RingSize(Feature):
    def call(self, mol: chem.Mol) -> list[int, float, str]: 
        def ring_size(atom):
            if not atom.IsInRing():
                return -1
            size = 3
            while not atom.IsInRingSize(size):
                size += 1 
            return size
        return [ring_size(atom) for atom in mol.atoms]
    

@keras.saving.register_keras_serializable(package='molcraft')
class FormalCharge(Feature):
    def call(self, mol: chem.Mol) -> list[int, float, str]: 
        return [atom.GetFormalCharge() for atom in mol.atoms]


@keras.saving.register_keras_serializable(package='molcraft')
class IsChiralityPossible(Feature):
    def call(self, mol: chem.Mol) -> list[int, float, str]: 
        return [atom.HasProp("_ChiralityPossible") for atom in mol.atoms]


@keras.saving.register_keras_serializable(package='molcraft')
class NumRadicalElectrons(Feature):
    def call(self, mol: chem.Mol) -> list[int, float, str]: 
        return [atom.GetNumRadicalElectrons() for atom in mol.atoms]


@keras.saving.register_keras_serializable(package='molcraft')
class IsAromatic(Feature):
    def call(self, mol: chem.Mol) -> list[int, float, str]: 
        return [atom.GetIsAromatic() for atom in mol.atoms]


@keras.saving.register_keras_serializable(package='molcraft')
class IsHeteroatom(Feature):
    def call(self, mol: chem.Mol) -> list[int, float, str]: 
        return chem.hetero_atoms(mol)


@keras.saving.register_keras_serializable(package='molcraft')
class IsHydrogenDonor(Feature):
    def call(self, mol: chem.Mol) -> list[int, float, str]: 
        return chem.hydrogen_donors(mol)


@keras.saving.register_keras_serializable(package='molcraft')
class IsHydrogenAcceptor(Feature):
    def call(self, mol: chem.Mol) -> list[int, float, str]: 
        return chem.hydrogen_acceptors(mol)


@keras.saving.register_keras_serializable(package='molcraft')
class IsInRing(Feature):
    def call(self, mol: chem.Mol) -> list[int, float, str]:         
        return [atom.IsInRing() for atom in mol.atoms]
    

@keras.saving.register_keras_serializable(package='molcraft')
class PartialCharge(Feature):
    """Gasteiger partial charge."""
    def call(self, mol: chem.Mol) -> list[int, float, str]:  
        return chem.partial_charges(mol)


@keras.saving.register_keras_serializable(package='molcraft')
class TotalPolarSurfaceAreaContribution(Feature):
    """Total polar surface area (TPSA) contribution."""
    def call(self, mol: chem.Mol) -> list[int, float, str]:      
        return chem.total_polar_surface_area_contributions(mol)


@keras.saving.register_keras_serializable(package='molcraft')
class AccessibleSurfaceAreaContribution(Feature):
    """Labute accessible surface area (ASA) contribution."""
    def call(self, mol: chem.Mol) -> list[int, float, str]:  
        return chem.accessible_surface_area_contributions(mol)


@keras.saving.register_keras_serializable(package='molcraft')
class LogPContribution(Feature):
    """Crippen logP contribution."""
    def call(self, mol: chem.Mol) -> list[int, float, str]:      
        return chem.logp_contributions(mol)


@keras.saving.register_keras_serializable(package='molcraft')
class MolarRefractivityContribution(Feature):
    """Crippen molar refractivity contribution."""
    def call(self, mol: chem.Mol) -> list[int, float, str]:      
        return chem.molar_refractivity_contributions(mol)


@keras.saving.register_keras_serializable(package='molcraft')
class BondType(Feature):
    def call(self, mol: chem.Mol) -> list[int, float, str]:  
        return [str(bond.GetBondType()).lower() for bond in mol.bonds]


@keras.saving.register_keras_serializable(package='molcraft')
class Stereo(Feature):
    def call(self, mol: chem.Mol) -> list[int, float, str]:  
        return [
            str(bond.GetStereo()).replace('STEREO', '').capitalize() 
            for bond in mol.bonds
        ]
    

@keras.saving.register_keras_serializable(package='molcraft')
class IsConjugated(Feature):
    def call(self, mol: chem.Mol) -> list[int, float, str]:  
        return [bond.GetIsConjugated() for bond in mol.bonds]


@keras.saving.register_keras_serializable(package='molcraft')
class IsRotatable(Feature):
    def call(self, mol: chem.Mol) -> list[int, float, str]:  
        return chem.rotatable_bonds(mol)


@keras.saving.register_keras_serializable(package='molcraft')
class PairFeature(Feature):

    def __call__(self, mol: chem.Mol) -> np.ndarray:
        if not isinstance(mol, chem.Mol):
            raise TypeError(f'Input to {self.name} must be a `chem.Mol` instance.')
        features = self.call(mol)
        if len(features) != int(mol.num_atoms**2):
            raise ValueError(
                f'The number of features computed by {self.name} does not '
                'match the number of node/atom pairs in the `chem.Mol` object. '
                f'Make sure the list of items returned by {self.name}(input) '
                'correspond to node/atom pairs: '
                '[(0, 0), (0, 1), ..., (0, N), (1, 0), ... (N, N)], '
                'where N denotes the number of nodes/atoms.'
            )
        func = (
            self._featurize_categorical if self.vocab else 
            self._featurize_floating
        )
        return np.asarray([func(x) for x in features], dtype=self.dtype)
        

@keras.saving.register_keras_serializable(package='molcraft')
class PairDistance(PairFeature):
    
    def __init__(
        self, 
        max_distance: int = None, 
        allow_oov: int = True,
        encode_oov: bool = True, 
        **kwargs,
    ) -> None:
        vocab = kwargs.pop('vocab', None)
        if not vocab:
            if max_distance is None:
                max_distance = 10
            vocab = list(range(max_distance + 1))
        super().__init__(
            vocab=vocab, 
            allow_oov=allow_oov, 
            encode_oov=encode_oov, 
            **kwargs
        )

    def call(self, mol: chem.Mol) -> list[int]:
        return [int(x) for x in chem.get_distances(mol).reshape(-1)]
    
    
default_vocabulary = {
    'AtomType': [
        '*', 'H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 
        'Mg', 'Al', 'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V', 
        'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 
        'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 
        'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I', 'Xe', 'Cs', 'Ba', 
        'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 
        'Er', 'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 
        'Au', 'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 
        'Th', 'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 
        'Md', 'No', 'Lr', 'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds', 'Rg', 
        'Cn', 'Nh', 'Fl', 'Mc', 'Lv', 'Ts', 'Og'
    ],
    'Degree': [
        0, 1, 2, 3, 4, 5, 6, 7, 8
    ],
    'TotalNumHs': [
        0, 1, 2, 3, 4
    ],
    'TotalValence': [
        0, 1, 2, 3, 4, 5, 6, 7, 8
    ],
    'Hybridization': [
        "s", "sp", "sp2", "sp3", "sp3d", "sp3d2", "unspecified"
    ],
    'CIPCode': [
        "R", "S", "None"
    ],
    'FormalCharge': [
        -3, -2, -1, 0, 1, 2, 3
    ],
    'NumRadicalElectrons': [
        0, 1, 2, 3, 4
    ],
    'RingSize': [
        -1, 3, 4, 5, 6, 7, 8
    ],
    'BondType': [
        "zero", "single", "double", "triple", "aromatic"
    ],
    'Stereo': [
        "E", "Z", "Any", "None"
    ],
}




import warnings
import keras
import numpy as np

from rdkit.Chem import rdMolDescriptors

from molcraft import chem
from molcraft import features


@keras.saving.register_keras_serializable(package='molcraft')
class Descriptor(features.Feature):

    def __call__(self, mol: chem.Mol) -> np.ndarray:
        if not isinstance(mol, chem.Mol):
            raise ValueError(
                f'Input to {self.name} must be a `chem.Mol` object.'
            )
        descriptor = self.call(mol)
        func = (
            self._featurize_categorical if self.vocab else 
            self._featurize_floating
        )
        if not isinstance(descriptor, (tuple, list, np.ndarray)):
            descriptor = [descriptor]
        
        descriptors = []
        for value in descriptor:
            descriptors.append(func(value))
        return np.concatenate(descriptors)
    

@keras.saving.register_keras_serializable(package='molcraft')
class Descriptor3D(Descriptor):
    
    def __call__(self, mol: chem.Mol) -> np.ndarray:
        if not isinstance(mol, chem.Mol):
            raise ValueError(
                f'Input to {self.name} must be a `chem.Mol` object.'
            ) 
        if mol.num_conformers == 0:
            raise ValueError(
                f'The inputted `chem.Mol` to {self.name} must embed a conformer. '
                f'It is recommended that {self.name} is used as a molecule feature '
                'for `MolGraphFeaturizer3D`, which by default embeds a conformer.'
            )
        return super().__call__(mol)


@keras.saving.register_keras_serializable(package='molcraft')
class MolWeight(Descriptor):
    def call(self, mol: chem.Mol) -> np.ndarray:
        return rdMolDescriptors.CalcExactMolWt(mol) 


@keras.saving.register_keras_serializable(package='molcraft')
class TotalPolarSurfaceArea(Descriptor):
    def call(self, mol: chem.Mol) -> np.ndarray:
        return rdMolDescriptors.CalcTPSA(mol)


@keras.saving.register_keras_serializable(package='molcraft')
class LogP(Descriptor):
    """Crippen logP."""
    def call(self, mol: chem.Mol) -> np.ndarray:
        return rdMolDescriptors.CalcCrippenDescriptors(mol)[0]
    

@keras.saving.register_keras_serializable(package='molcraft')
class MolarRefractivity(Descriptor):
    """Crippen molar refractivity."""
    def call(self, mol: chem.Mol) -> np.ndarray:
        return rdMolDescriptors.CalcCrippenDescriptors(mol)[1]
    

@keras.saving.register_keras_serializable(package='molcraft')
class NumAtoms(Descriptor):
    def call(self, mol: chem.Mol) -> np.ndarray:
        return rdMolDescriptors.CalcNumAtoms(mol)
    

@keras.saving.register_keras_serializable(package='molcraft')
class NumHeavyAtoms(Descriptor):
    def call(self, mol: chem.Mol) -> np.ndarray:
        return rdMolDescriptors.CalcNumHeavyAtoms(mol)


@keras.saving.register_keras_serializable(package='molcraft')
class NumHeteroatoms(Descriptor):
    def call(self, mol: chem.Mol) -> np.ndarray:
        return rdMolDescriptors.CalcNumHeteroatoms(mol) 
    
    
@keras.saving.register_keras_serializable(package='molcraft')
class NumHydrogenDonors(Descriptor):
    def call(self, mol: chem.Mol) -> np.ndarray:
        return rdMolDescriptors.CalcNumHBD(mol) 


@keras.saving.register_keras_serializable(package='molcraft')
class NumHydrogenAcceptors(Descriptor):
    def call(self, mol: chem.Mol) -> np.ndarray:
        return rdMolDescriptors.CalcNumHBA(mol)
    

@keras.saving.register_keras_serializable(package='molcraft')
class NumRotatableBonds(Descriptor):
    def call(self, mol: chem.Mol) -> np.ndarray:
        return rdMolDescriptors.CalcNumRotatableBonds(mol) 


@keras.saving.register_keras_serializable(package='molcraft')
class NumRings(Descriptor):
    def call(self, mol: chem.Mol) -> np.ndarray:
        return rdMolDescriptors.CalcNumRings(mol)


@keras.saving.register_keras_serializable(package='molcraft')
class NumAromaticRings(Descriptor):
    def call(self, mol: chem.Mol) -> np.ndarray:
        return rdMolDescriptors.CalcNumAromaticRings(mol)


@keras.saving.register_keras_serializable(package='molcraft')
class AtomCount(Descriptor):

    def __init__(self, atom_type: str, **kwargs):
        super().__init__(**kwargs)
        self.atom_type = atom_type

    def call(self, mol: chem.Mol) -> np.ndarray:
        count = 0
        for atom in mol.atoms:
            if atom.GetSymbol() == self.atom_type:
                count += 1
        return count
    
    def get_config(self) -> dict:
        config = super().get_config()
        config['atom_type'] = self.atom_type
        return config
    

@keras.saving.register_keras_serializable(package='molcraft')
class ForceFieldEnergy(Descriptor3D):
    """Universal Force Field (UFF) Energy."""
    def call(self, mol: chem.Mol) -> np.ndarray:
        return chem.conformer_energies(mol, method="UFF")
        


import unittest 
import tempfile
import shutil
import keras
import numpy as np

from molcraft import tensors 
from molcraft import layers 
from molcraft import models


class TestModel(unittest.TestCase):

    def setUp(self):

        self.tensors = [
            # Graph with two subgraphs
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([2, 3], dtype='int32'),
                    'label': keras.ops.array([1., 2.], dtype='float32'),
                    'weight': keras.ops.array([0.5, 1.25], dtype='float32'),
                },
                node={
                    'feature': keras.ops.array([[1., 2.], [3., 4.], [5., 6.], [6., 7.], [8., 9.]], dtype='float32'),
                    'weight': keras.ops.array([0.50, 1.00, 2.00, 0.25, 0.75], dtype='float32')
                },
                edge={
                    'source': keras.ops.array([0, 1, 3, 4, 4, 3, 2], dtype='int32'),
                    'target': keras.ops.array([1, 0, 2, 3, 4, 4, 3], dtype='int32'),
                    'feature': keras.ops.array([[1.], [2.], [3.], [4.], [5.], [6.], [7.]], dtype='float32')
                }
            ),
            # Graph with two subgraphs, none which has edges
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([2, 3], dtype='int32'),
                    'label': keras.ops.array([1., 2.], dtype='float32'),
                    'weight': keras.ops.array([0.5, 1.25], dtype='float32'),
                },
                node={
                    'feature': keras.ops.array([[1., 2.], [3., 4.], [5., 6.], [6., 7.], [8., 9.]], dtype='float32'),
                    'weight': keras.ops.array([0.50, 1.00, 2.00, 0.25, 0.75], dtype='float32')
                },
                edge={
                    'source': keras.ops.zeros([0], dtype='int32'),
                    'target': keras.ops.zeros([0], dtype='int32'),
                    'feature': keras.ops.zeros([0, 1], dtype='float32')
                }
            ),
            # Graph with two subgraphs, none of which has nodes or edges
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([0, 0], dtype='int32'),
                    'label': keras.ops.array([1., 2.], dtype='float32'),
                    'weight': keras.ops.array([0.5, 1.25], dtype='float32'),
                },
                node={
                    'feature': keras.ops.zeros([0, 2], dtype='float32'),
                    'weight': keras.ops.zeros([0], dtype='float32')
                },
                edge={
                    'source': keras.ops.zeros([0], dtype='int32'),
                    'target': keras.ops.zeros([0], dtype='int32'),
                    'feature': keras.ops.zeros([0, 1], dtype='float32')
                }
            ),
            # Graph with three subgraphs where the second subgraph's first node has no edges
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([2, 3, 2], dtype='int32'),
                    'label': keras.ops.array([1., 2., 3.], dtype='float32'),
                    'weight': keras.ops.array([0.5, 1.25, 0.75], dtype='float32'),
                },
                node={
                    'feature': keras.ops.array([[1., 2.], [3., 4.], [5., 6.], [6., 7.], [8., 9.], [10., 11.], [12., 13.]], dtype='float32'),
                    'weight': keras.ops.array([0.50, 1.00, 2.00, 0.25, 0.75, 0.25, 0.25], dtype='float32')
                },
                edge={
                    'source': keras.ops.array([0, 1, 4, 3, 4, 5, 6, 6], dtype='int32'),
                    'target': keras.ops.array([1, 0, 3, 4, 4, 6, 5, 6], dtype='int32'),
                    'feature': keras.ops.array([[1.], [2.], [3.], [4.], [5.], [6.], [7.], [8.]], dtype='float32')
                }
            ),
            # Graph with three subgraphs where the first subgraph's nodes have no edges
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([2, 3, 2], dtype='int32'),
                    'label': keras.ops.array([1., 2., 3.], dtype='float32'),
                    'weight': keras.ops.array([0.5, 1.25, 0.75], dtype='float32'),
                },
                node={
                    'feature': keras.ops.array([[1., 2.], [3., 4.], [5., 6.], [6., 7.], [8., 9.], [10., 11.], [12., 13.]], dtype='float32'),
                    'weight': keras.ops.array([0.50, 1.00, 2.00, 0.25, 0.75, 0.25, 0.25], dtype='float32')
                },
                edge={
                    'source': keras.ops.array([4, 3, 4, 5, 6, 6], dtype='int32'),
                    'target': keras.ops.array([3, 4, 4, 6, 5, 6], dtype='int32'),
                    'feature': keras.ops.array([[3.], [4.], [5.], [6.], [7.], [8.]], dtype='float32')
                }
            ),
            # Graph with three subgraphs where the last subgraph's nodes have no edges
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([2, 3, 3], dtype='int32'),
                    'label': keras.ops.array([1., 2., 3.], dtype='float32'),
                    'weight': keras.ops.array([0.5, 1.25, 0.75], dtype='float32'),
                },
                node={
                    'feature': keras.ops.array([[1., 2.], [3., 4.], [5., 6.], [6., 7.], [8., 9.], [10., 11.], [12., 13.], [14., 15.]], dtype='float32'),
                    'weight': keras.ops.array([0.50, 1.00, 2.00, 0.25, 0.75, 0.25, 0.25, 0.5], dtype='float32')
                },
                edge={
                    'source': keras.ops.array([0, 1, 4, 3, 4, 5, 6], dtype='int32'),
                    'target': keras.ops.array([1, 0, 3, 4, 4, 6, 5], dtype='int32'),
                    'feature': keras.ops.array([[1.], [2.], [3.], [4.], [5.], [6.], [7.]], dtype='float32')
                }
            ),
            # Graph with two subgraphs with wildcards
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([2, 3], dtype='int32'),
                    'label': keras.ops.array([1., 2.], dtype='float32'),
                    'weight': keras.ops.array([0.5, 1.25], dtype='float32'),
                },
                node={
                    'feature': keras.ops.array([[1., 2.], [3., 4.], [5., 6.], [6., 7.], [8., 9.]], dtype='float32'),
                    'weight': keras.ops.array([0.50, 1.00, 2.00, 0.25, 0.75], dtype='float32'),
                    "wildcard": keras.ops.array([1, 2, 1, 0, 1]),
                },
                edge={
                    'source': keras.ops.array([0, 1, 3, 4, 4, 3, 2], dtype='int32'),
                    'target': keras.ops.array([1, 0, 2, 3, 4, 4, 3], dtype='int32'),
                    'feature': keras.ops.array([[1.], [2.], [3.], [4.], [5.], [6.], [7.]], dtype='float32')
                }
            ),
        ]

    def test_functional_model(self):

        def get_model(tensor):
            inputs = layers.Input(tensor.spec)
            x = layers.NodeEmbedding(32)(inputs)
            x = layers.EdgeEmbedding(32)(x)
            x = layers.GTConv(32)(x)
            x = layers.GTConv(32)(x)
            x = layers.Readout('sum')(x)
            outputs = keras.layers.Dense(1)(x)
            return models.GraphModel(inputs, outputs)
        
        for i, tensor in enumerate(self.tensors):
            with self.subTest(i=i, functional=True):
                model = get_model(tensor)
                output = model(tensor)
                self.assertTrue(output.shape[0] == tensor.context['label'].shape[0])
                model.compile('adam', 'mse', metrics=[keras.metrics.MeanAbsoluteError()])
                metrics = model.evaluate(tensor, verbose=0)
                self.assertTrue(isinstance(metrics, list))
                del model

    def test_saved_model(self):

        def get_model(tensor):
            inputs = layers.Input(tensor.spec)
            x = layers.NodeEmbedding(32)(inputs)
            x = layers.EdgeEmbedding(32)(x)
            x = layers.GTConv(32)(x)
            x = layers.GTConv(32)(x)
            x = layers.Readout('sum')(x)
            outputs = keras.layers.Dense(1)(x)
            return models.GraphModel(inputs, outputs)

        @keras.saving.register_keras_serializable()
        class Model(models.GraphModel):
            def __init__(self, **kwargs):
                super().__init__(**kwargs)
                self.e1 = layers.NodeEmbedding(32)
                self.e2 = layers.EdgeEmbedding(32)
                self.c1 = layers.GTConv(32)
                self.c2 = layers.GTConv(32)
                self.r = layers.Readout('sum')
                self.d = keras.layers.Dense(1)
            def propagate(self, tensor):
                return self.d(self.r(self.c2(self.c1(self.e2(self.e1(tensor))))))
            
        example = self.tensors[-1]

        tmp_dir = tempfile.mkdtemp()
        tmp_file = tmp_dir + '/model.keras'
        with self.subTest(functional_model=True):
            model = get_model(example)
            model.compile('adam', 'mse')
            model.fit(example, verbose=0)
            model.save(tmp_file)
            loaded_model = models.load_model(tmp_file)
            pred_1 = model.predict(example, verbose=0)
            pred_2 = loaded_model.predict(example, verbose=0)
            test_preds = np.all(pred_1.round(4) == pred_2.round(4))
            self.assertTrue(test_preds)
            test_vars = np.all([np.all((v1 == v2).numpy()) for (v1, v2) in zip(model.variables, loaded_model.variables)])
            self.assertTrue(test_vars)
        shutil.rmtree(tmp_dir)

        tmp_dir = tempfile.mkdtemp()
        tmp_file = tmp_dir + '/model.keras'
        with self.subTest(functional_model=False):
            model = Model()
            model.compile('adam', 'mse')
            model.fit(example, verbose=0)
            model.save(tmp_file)
            loaded_model = models.load_model(tmp_file)
            pred_1 = model.predict(example, verbose=0)
            pred_2 = loaded_model.predict(example, verbose=0)
            test_preds = np.all(pred_1.round(4) == pred_2.round(4))
            self.assertTrue(test_preds)
            test_vars = np.all([np.all((v1 == v2).numpy()) for (v1, v2) in zip(model.variables, loaded_model.variables)])
            self.assertTrue(test_vars)
        shutil.rmtree(tmp_dir)

    def test_subclassed_model(self):

        def get_model(tensor):
            class Model(models.GraphModel):
                def __init__(self, **kwargs):
                    super().__init__(**kwargs)
                    self.e1 = layers.NodeEmbedding(32)
                    self.e2 = layers.EdgeEmbedding(32)
                    self.c1 = layers.GTConv(32)
                    self.c2 = layers.GTConv(32)
                    self.r = layers.Readout('sum')
                    self.d = keras.layers.Dense(1)
                def propagate(self, tensor):
                    return self.d(self.r(self.c2(self.c1(self.e2(self.e1(tensor))))))
            return Model()
                
        for i, tensor in enumerate(self.tensors):
            with self.subTest(i=i, functional=True):
                model = get_model(tensor)
                output = model(tensor)
                self.assertTrue(output.shape[0] == tensor.context['label'].shape[0])
                model.compile('adam', 'mse', metrics=[keras.metrics.MeanAbsoluteError()])
                metrics = model.evaluate(tensor, verbose=0)
                self.assertTrue(isinstance(metrics, list))
                del model

    def test_interpet(self):

        def get_model(tensor):
            inputs = layers.Input(tensor.spec)
            x = layers.NodeEmbedding(32)(inputs)
            x = layers.EdgeEmbedding(32)(x)
            x = layers.GTConv(32)(x)
            x = layers.GTConv(32)(x)
            x = layers.Readout('sum')(x)
            outputs = keras.layers.Dense(1)(x)
            return models.GraphModel(inputs, outputs)
        
        for i, tensor in enumerate(self.tensors):
            with self.subTest(i=i, functional=True):
                model = get_model(tensor)
                tensor = models.interpret(model, tensor)
                self.assertTrue('saliency' in tensor.node)

                tensor = tensor.update({'node': {'saliency': None}})
                tensor = models.saliency(model, tensor)
                self.assertTrue('feature_saliency' in tensor.node)

    def test_embedding(self):

        units = 32

        def get_model(tensor):
            inputs = layers.Input(tensor.spec)
            x = layers.NodeEmbedding(units)(inputs)
            x = layers.EdgeEmbedding(units)(x)
            x = layers.GTConv(units)(x)
            x = layers.GTConv(units)(x)
            x = layers.Readout('sum')(x)
            outputs = keras.layers.Dense(1)(x)
            return models.GraphModel(inputs, outputs)
        
        for i, tensor in enumerate(self.tensors):
            with self.subTest(i=i, functional=True):
                model = get_model(tensor)
                out = model.embedding()(tensor)
                self.assertTrue(out.shape[0] == tensor.context['size'].shape[0])
                self.assertTrue(out.shape[1] == units)


if __name__ == '__main__':
    unittest.main()